# -*- coding: utf-8 -*-
# GTXStorage Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2017 GTXStorage, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
gtxstorage.thread_pool
~~~~~~~~~~~~

This module implements a thread pool API to run several tasks
in parallel. Tasks results can also be retrieved.

:copyright: (c) 2017 by GTXStorage, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import sys
from threading import Thread
from .compat import queue

class Worker(Thread):
    """ Thread executing tasks from a given tasks queue """
    def __init__(self, tasks_queue, results_queue, exceptions_queue):
        Thread.__init__(self)
        self.tasks_queue = tasks_queue
        self.results_queue = results_queue
        self.exceptions_queue = exceptions_queue
        self.daemon = True
        self.start()

    def run(self):
        fast_quit = False
        while True:
            func, args, kargs = self.tasks_queue.get()
            if not fast_quit:
                try:
                    result = func(*args, **kargs)
                    self.results_queue.put(result)
                except Exception as e:
                    self.exceptions_queue.put(e)
                    fast_quit = True
            # Mark this task as done, whether an exception happened or not
            self.tasks_queue.task_done()


class ThreadPool:
    """ Pool of threads consuming tasks from a queue """
    def __init__(self, num_threads):
        self.results_queue = queue()
        self.exceptions_queue = queue()
        self.tasks_queue = queue(num_threads)
        for _ in range(num_threads):
            Worker(self.tasks_queue, self.results_queue, self.exceptions_queue)

    def add_task(self, func, *args, **kargs):
        """ Add a task to the queue """
        self.tasks_queue.put((func, args, kargs))

    def parallel_run(self, func, args_list):
        """ Add a list of tasks to the queue """
        for args in args_list:
            self.add_task(func, args)
        # Wait for completion of all the tasks in the queue
        self.tasks_queue.join()
        # Check if one of the thread raised an exception, if yes
        # raise it here in the function
        if not self.exceptions_queue.empty():
            raise self.exceptions_queue.get()

    def result(self):
        """ Return the result of all called tasks """
        return self.results_queue




# -*- coding: utf-8 -*-
# GTXStorage Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016, 2017 GTXStorage, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
gtxstorage.helpers

This module implements all helper functions.

:copyright: (c) 2015, 2016, 2017 by GTXStorage, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

from __future__ import absolute_import
import io

import collections
import base64
import hashlib
import re
import os
import errno
import math

from .compat import (urlsplit, urlencode, queryencode,
                     str, bytes, basestring)
from .error import (InvalidBucketError, InvalidEndpointError,
                    InvalidArgumentError)

# Constants
MAX_MULTIPART_COUNT = 10000 # 10000 parts
MAX_MULTIPART_OBJECT_SIZE = 5 * 1024 * 1024 * 1024 * 1024  # 5TiB
MIN_PART_SIZE = 5 * 1024 * 1024  # 5MiB

_VALID_BUCKETNAME_REGEX = re.compile('^[a-z0-9][a-z0-9\\.\\-]+[a-z0-9]$')
_ALLOWED_HOSTNAME_REGEX = re.compile(
    '^((?!-)[A-Z\\d-]{1,63}(?<!-)\\.)*((?!-)[A-Z\\d-]{1,63}(?<!-))$',
    re.IGNORECASE)

_EXTRACT_REGION_REGEX = re.compile('s3[.-]?(.+?).amazonaws.com')

def get_s3_region_from_endpoint(endpoint):
    """
    Extracts and returns an AWS S3 region from an endpoint
    of form `s3-ap-southeast-1.amazonaws.com`

    :param endpoint: Endpoint region to be extracted.
    """

    # Extract region by regex search.
    m = _EXTRACT_REGION_REGEX.search(endpoint)
    if m:
        # Regex matches, we have found a region.
        region = m.group(1)
        if region == 'external-1':
            # Handle special scenario for us-east-1 URL.
            return 'us-east-1'
        if region.startswith('dualstack'):
            # Handle special scenario for dualstack URL.
            return region.split('.')[1]
        return region

    # No regex matches return None.
    return None

def dump_http(method, url, request_headers, response, output_stream):
    """
    Dump all headers and response headers into output_stream.

    :param request_headers: Dictionary of HTTP request headers.
    :param response_headers: Dictionary of HTTP response headers.
    :param output_stream: Stream where the request is being dumped at.
    """

    # Start header.
    output_stream.write('---------START-HTTP---------\n')

    # Get parsed url.
    parsed_url = urlsplit(url)

    # Dump all request headers recursively.
    http_path = parsed_url.path
    if parsed_url.query:
        http_path = http_path + '?' + parsed_url.query

    output_stream.write('{0} {1} HTTP/1.1\n'.format(method,
                                                    http_path))

    for k, v in list(request_headers.items()):
        if k is 'authorization':
            # Redact signature header value from trace logs.
            v = re.sub(r'Signature=([[0-9a-f]+)', 'Signature=*REDACTED*', v)
        output_stream.write('{0}: {1}\n'.format(k.title(), v))

    # Write a new line.
    output_stream.write('\n')

    # Write response status code.
    output_stream.write('HTTP/1.1 {0}\n'.format(response.status))

    # Dump all response headers recursively.
    for k, v in list(response.getheaders().items()):
        output_stream.write('{0}: {1}\n'.format(k.title(), v))

    # For all errors write all the available response body.
    if response.status != 200 and \
       response.status != 204 and response.status != 206:
        output_stream.write('{0}'.format(response.read()))

    # End header.
    output_stream.write('---------END-HTTP---------\n')

def mkdir_p(path):
    """
    Recursively creates parent and sub directories.

    :param path:
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

class PartMetadata(object):
    """
    Parts manager split parts metadata :class:`PartMetadata <PartMetadata>`.

    :param data: Part writer object backed by temporary file.
    :param md5_hex: MD5 hash in hex format.
    :param sha256_hex: Sha256 hash in hex format.
    :param size: Size of the part.
    """
    def __init__(self, data, md5_hex, sha256_hex, size):
        self.data = data
        self.md5_hex = md5_hex
        self.sha256_hex = sha256_hex
        self.size = size

def read_full(data, size):
    """
    read_full reads exactly `size` bytes from reader. returns
    `size` bytes.

    :param data: Input stream to read from.
    :param size: Number of bytes to read from `data`.
    :return: Returns :bytes:`part_data`
    """
    default_read_size = 32768 # 32KiB per read operation.
    chunk = io.BytesIO()
    chunk_size = 0

    while chunk_size < size:
        read_size = default_read_size
        if (size - chunk_size) < default_read_size:
            read_size = size - chunk_size
        current_data = data.read(read_size)
        if not current_data or len(current_data) == 0:
            break
        chunk.write(current_data)
        chunk_size+= len(current_data)

    return chunk.getvalue()

AWS_S3_ENDPOINT_MAP = {
    'us-east-1': 's3.amazonaws.com',
    'us-east-2': 's3-us-east-2.amazonaws.com',
    'us-west-2': 's3-us-west-2.amazonaws.com',
    'us-west-1': 's3-us-west-1.amazonaws.com',
    'ca-central-1': 's3.ca-central-1.amazonaws.com',
    'eu-west-1': 's3-eu-west-1.amazonaws.com',
    'eu-west-2': 's3-eu-west-2.amazonaws.com',
    'sa-east-1': 's3-sa-east-1.amazonaws.com',
    'eu-central-1': 's3-eu-central-1.amazonaws.com',
    'ap-south-1': 's3.ap-south-1.amazonaws.com',
    'ap-southeast-1': 's3-ap-southeast-1.amazonaws.com',
    'ap-northeast-1': 's3-ap-northeast-1.amazonaws.com',
    'ap-northeast-2': 's3-ap-northeast-2.amazonaws.com',
    'cn-north-1': 's3.cn-north-1.amazonaws.com.cn'
}

def get_s3_endpoint(region):
    return AWS_S3_ENDPOINT_MAP.get(region, 's3.amazonaws.com')

def get_target_url(endpoint_url, bucket_name=None, object_name=None,
                   bucket_region='us-east-1', query=None):
    """
    Construct final target url.

    :param endpoint_url: Target endpoint url where request is served to.
    :param bucket_name: Bucket component for the target url.
    :param object_name: Object component for the target url.
    :param bucket_region: Bucket region for the target url.
    :param query: Query parameters as a *dict* for the target url.
    :return: Returns final target url as *str*.
    """
    # New url
    url = None

    # Parse url
    parsed_url = urlsplit(endpoint_url)

    # Get new host, scheme.
    scheme = parsed_url.scheme
    host = parsed_url.netloc

    # Strip 80/443 ports since curl & browsers do not
    # send them in Host header.
    if parsed_url.port == 80 or parsed_url.port == 443:
        host = parsed_url.hostname

    if 's3.amazonaws.com' in host:
        host = get_s3_endpoint(bucket_region)

    url = scheme + '://' + host
    if bucket_name:
        # Save if target url will have buckets which suppport
        # virtual host.
        is_virtual_host_style = is_virtual_host(endpoint_url,
                                                bucket_name)
        if is_virtual_host_style:
            url = (scheme + '://' + bucket_name + '.' + host)
        else:
            url = (scheme + '://' + host + '/' + bucket_name)

    url_components = [url]
    url_components.append('/')

    if object_name:
        object_name = encode_object_name(object_name)
        url_components.append(object_name)

    if query:
        ordered_query = collections.OrderedDict(sorted(query.items()))
        query_components = []
        for component_key in ordered_query:
            if ordered_query[component_key] is not None:
                if isinstance(ordered_query[component_key], list):
                    for value in ordered_query[component_key]:
                        query_components.append(component_key+'='+
                                                queryencode(value))
                else:
                    query_components.append(
                        component_key+'='+
                        queryencode(
                            ordered_query[component_key]
                        )
                    )
            else:
                query_components.append(component_key)

        query_string = '&'.join(query_components)
        if query_string:
            url_components.append('?')
            url_components.append(query_string)

    return ''.join(url_components)


def is_valid_endpoint(endpoint):
    """
    Verify if endpoint is valid.

    :type endpoint: string
    :param endpoint: An endpoint. Must have at least a scheme and a hostname.
    :return: True if the endpoint is valid. Raise :exc:`InvalidEndpointError`
       otherwise.
    """
    try:
        if urlsplit(endpoint).scheme:
            raise InvalidEndpointError('Hostname cannot have a scheme.')

        hostname = endpoint.split(':')[0]
        if hostname is None:
            raise InvalidEndpointError('Hostname cannot be empty.')

        if len(hostname) > 255:
            raise InvalidEndpointError('Hostname cannot be greater than 255.')

        if hostname[-1] == '.':
            hostname = hostname[:-1]

        if not _ALLOWED_HOSTNAME_REGEX.match(hostname):
            raise InvalidEndpointError('Hostname does not meet URL standards.')

    except AttributeError as error:
        raise TypeError(error)

    return True

def is_virtual_host(endpoint_url, bucket_name):
    """
    Check to see if the ``bucket_name`` can be part of virtual host
    style.

    :param endpoint_url: Endpoint url which will be used for virtual host.
    :param bucket_name: Bucket name to be validated against.
    """
    is_valid_bucket_name(bucket_name)

    parsed_url = urlsplit(endpoint_url)
    # bucket_name can be valid but '.' in the hostname will fail
    # SSL certificate validation. So do not use host-style for
    # such buckets.
    if 'https' in parsed_url.scheme and '.' in bucket_name:
        return False
    if 's3.amazonaws.com' in parsed_url.netloc:
        return True
    return False

def is_valid_bucket_name(bucket_name):
    """
    Check to see if the ``bucket_name`` complies with the
    restricted DNS naming conventions necessary to allow
    access via virtual-hosting style.

    :param bucket_name: Bucket name in *str*.
    :return: True if the bucket is valid. Raise :exc:`InvalidBucketError`
       otherwise.
    """
    # Verify bucket name length.
    if len(bucket_name) < 3:
        raise InvalidBucketError('Bucket name cannot be less than'
                                 ' 3 characters.')
    if len(bucket_name) > 63:
        raise InvalidBucketError('Bucket name cannot be more than'
                                 ' 63 characters.')
    if '..' in bucket_name:
        raise InvalidBucketError('Bucket name cannot have successive'
                                 ' periods.')

    match = _VALID_BUCKETNAME_REGEX.match(bucket_name)
    if match is None or match.end() != len(bucket_name):
        raise InvalidBucketError('Bucket name does not follow S3 standards.'
                                 ' Bucket: {0}'.format(bucket_name))

    return True


def is_non_empty_string(input_string):
    """
    Validate if non empty string

    :param input_string: Input is a *str*.
    :return: True if input is string and non empty.
       Raise :exc:`Exception` otherwise.
    """
    try:
        if not input_string.strip():
            raise ValueError()
    except AttributeError as error:
        raise TypeError(error)

    return True


def is_valid_bucket_notification_config(notifications):
    """
    Validate the notifications config structure

    :param notifications: Dictionary with specific structure.
    :return: True if input is a valid bucket notifications structure.
       Raise :exc:`InvalidArgumentError` otherwise.
    """
    # check if notifications is a dict.
    if not isinstance(notifications, dict):
        raise TypeError('notifications configuration must be a dictionary')

    if len(notifications) == 0:
        raise InvalidArgumentError(
            'notifications configuration may not be empty'
        )

    VALID_NOTIFICATION_KEYS = set([
        "TopicConfigurations",
        "QueueConfigurations",
        "CloudFunctionConfigurations",
    ])

    VALID_SERVICE_CONFIG_KEYS = set([
        'Id',
        'Arn',
        'Events',
        'Filter',
    ])

    NOTIFICATION_EVENTS = set([
        's3:ReducedRedundancyLostObject',
        's3:ObjectCreated:*',
        's3:ObjectCreated:Put',
        's3:ObjectCreated:Post',
        's3:ObjectCreated:Copy',
        's3:ObjectCreated:CompleteMultipartUpload',
        's3:ObjectRemoved:*',
        's3:ObjectRemoved:Delete',
        's3:ObjectRemoved:DeleteMarkerCreated',
    ])

    for key, value in notifications.items():
        # check if key names are valid
        if key not in VALID_NOTIFICATION_KEYS:
            raise InvalidArgumentError((
                '{} is an invalid key '
                'for notifications configuration').format(key))

        # check if config values conform
        # first check if value is a list
        if not isinstance(value, list):
            raise InvalidArgumentError((
                'The value for key "{}" in the notifications '
                'configuration must be a list.').format(key))

        for service_config in value:
            # check type matches
            if not isinstance(service_config, dict):
                raise InvalidArgumentError((
                    'Each service configuration item for "{}" must be a '
                    'dictionary').format(key))

            # check keys are valid
            for skey in service_config.keys():
                if skey not in VALID_SERVICE_CONFIG_KEYS:
                    raise InvalidArgumentError((
                        '{} is an invalid key for a service '
                        'configuration item').format(skey))

            # check for required keys
            arn = service_config.get('Arn', '')
            if arn == '':
                raise InvalidArgumentError(
                    'Arn key in service config must be present and has to be '
                    'non-empty string'
                )
            events = service_config.get('Events', [])
            if len(events) < 1:
                raise InvalidArgumentError(
                    'At least one event must be specified in a service config'
                )
            if not isinstance(events, list):
                raise InvalidArgumentError('"Events" must be a list of strings '
                                           'in a service configuration')

            # check if 'Id' key is present, it should be string or bytes.
            if not isinstance(service_config.get('Id', ''), basestring):
                raise InvalidArgumentError('"Id" key must be a string')

            for event in events:
                if event not in NOTIFICATION_EVENTS:
                    raise InvalidArgumentError(
                        '{} is not a valid event. Valid '
                        'events are: {}'.format(event, NOTIFICATION_EVENTS))

            if 'Filter' in service_config:
                exception_msg = (
                    '{} - If a Filter key is given, it must be a '
                    'dictionary, the dictionary must have the '
                    'key "Key", and its value must be an object, with '
                    'a key named "FilterRules" which must be a non-empty list.'
                ).format(
                    service_config['Filter']
                )
                try:
                    filter_rules = service_config.get('Filter', {}).get(
                        'Key', {}).get('FilterRules', [])
                    if not isinstance(filter_rules, list):
                        raise InvalidArgumentError(exception_msg)
                    if len(filter_rules) < 1:
                        raise InvalidArgumentError(exception_msg)
                except AttributeError:
                    raise InvalidArgumentError(exception_msg)
                for filter_rule in filter_rules:
                    try:
                        name = filter_rule['Name']
                        value = filter_rule['Value']
                    except KeyError:
                        raise InvalidArgumentError((
                            '{} - a FilterRule dictionary must have "Name" '
                             'and "Value" keys').format(filter_rule))

                    if name not in ['prefix', 'suffix']:
                        raise InvalidArgumentError((
                            '{} - The "Name" key in a filter rule must be '
                             'either "prefix" or "suffix"').format(name))

    return True


def encode_object_name(object_name):
    """
    URL encode input object name.

    :param object_name: Un-encoded object name.
    :return: URL encoded input object name.
    """
    is_non_empty_string(object_name)
    return urlencode(object_name)

class Hasher(object):
    """
    Adaptation of hashlib-based hash functions that
    return unicode-encoded hex- and base64-digest
    strings.
    """
    def __init__(self, data, h):
        if data is None:
            data = b''
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.h = h(data)

    @classmethod
    def md5(cls, data=''):
        return cls(data, hashlib.md5)

    @classmethod
    def sha256(cls, data=''):
        return cls(data, hashlib.sha256)

    def update(self, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.h.update(data)

    def hexdigest(self):
        r = self.h.hexdigest()
        return r.decode('utf-8') if isinstance(r, bytes) else r

    def base64digest(self):
        r = base64.b64encode(self.h.digest())
        return r.decode('utf-8') if isinstance(r, bytes) else r


def get_sha256_hexdigest(content):
    """
    Calculate sha256 hexdigest of content.

    :param content: Input str or bytes. If the type is `str` we encode
    it to UTF8 first.

    :return: sha256 digest encoded as hexadecimal `str`.

    """
    return Hasher.sha256(content).hexdigest()


def get_md5_base64digest(content):
    """Calculate md5sum and return digest as base64 encoded string.

    :param content: Input str or bytes. If the type is `str` we encode
    it to UTF8 and calculate md5sum.

    :return: md5 digest encoded to base64 `str`.

    """
    return Hasher.md5(content).base64digest()


def optimal_part_info(length):
    """
    Calculate optimal part size for multipart uploads.

    :param length: Input length to calculate part size of.
    :return: Optimal part size.
    """
    # object size is '-1' set it to 5TiB.
    if length == -1:
        length = MAX_MULTIPART_OBJECT_SIZE
    if length > MAX_MULTIPART_OBJECT_SIZE:
        raise InvalidArgumentError('Input content size is bigger '
                                   ' than allowed maximum of 5TiB.')

    # Use floats for part size for all calculations to avoid
    # overflows during float64 to int64 conversions.
    part_size_float = math.ceil(length/MAX_MULTIPART_COUNT)
    part_size_float = (math.ceil(part_size_float/MIN_PART_SIZE)
                       * MIN_PART_SIZE)
    # Total parts count.
    total_parts_count = int(math.ceil(length/part_size_float))
    # Part size.
    part_size = int(part_size_float)
    # Last part size.
    last_part_size = length - int(total_parts_count-1)*part_size
    return total_parts_count, part_size, last_part_size



# -*- coding: utf-8 -*-
# GTXStorage Python Library for Amazon S3 Compatible Cloud Storage, (C)
# 2015, 2016 GTXStorage, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
gtxstorage.policy

This module implements policy management.

:copyright: (c) 2016 by GTXStorage, Inc.
:license: Apache 2.0, see LICENSE for more details.

"""

import collections
import fnmatch
import itertools

from .compat import basestring

_AWS_RESOURCE_PREFIX = "arn:aws:s3:::"

_COMMON_BUCKET_ACTIONS = set(['s3:GetBucketLocation'])
_READ_ONLY_BUCKET_ACTIONS = set(['s3:ListBucket'])
_WRITE_ONLY_BUCKET_ACTIONS = set(['s3:ListBucketMultipartUploads'])

_READ_ONLY_OBJECT_ACTIONS = set(['s3:GetObject'])
_WRITE_ONLY_OBJECT_ACTIONS = set(['s3:AbortMultipartUpload',
                                  's3:DeleteObject',
                                  's3:ListMultipartUploadParts',
                                  's3:PutObject'])
_READ_WRITE_OBJECT_ACTIONS = (_READ_ONLY_OBJECT_ACTIONS |
                              _WRITE_ONLY_OBJECT_ACTIONS)


# Policy Enums.
class Policy(object):
    NONE = "none"
    READ_ONLY = "readonly"
    READ_WRITE = "readwrite"
    WRITE_ONLY = "writeonly"


def _get_resource(statement):
    """
    Always return resource as a list
    """
    resource = statement.get('Resource')
    if isinstance(resource, basestring):
        resource = [resource]
    return resource


def _get_action(statement):
    """
    Always return action as a list
    """
    action = statement.get('Action', [])
    if isinstance(action, basestring):
        action = [action]
    return action


def _get_bucket_resource(bucket_name):
    """
    :param bucket_name: Name of the bucket
    :type bucket_name: str

    :return: Representation of the bucket with the resource prefix
    :rtype: str
    """
    return [_AWS_RESOURCE_PREFIX + bucket_name]


def _get_resource_prefix(bucket_name):
    """
    :param bucket_name: Name of the bucket
    :type bucket_name: str

    :return: Representation of the bucket with the resource prefix,
             but with a trailing slash.
    :rtype: str
    """
    return _get_bucket_resource(bucket_name)[0] + '/'


def _get_object_resource(bucket_name, prefix):
    """
    :param bucket_name: Name of the bucket
    :type bucket_name: str
    :param prefix: Name of the prefix
    :type prefix: str

    :return: Representation of an object in a bucket with the resource prefix.
    :rtype: str
    """
    return [_get_resource_prefix(bucket_name) + prefix + '*']


# Returns new statements with bucket actions.
def _new_bucket_statement(policy, bucket_name, prefix=''):
    if policy == Policy.NONE:
        return []

    bucket_resource = _get_bucket_resource(bucket_name)

    rv = [{'Action': list(_COMMON_BUCKET_ACTIONS),
           'Effect': 'Allow',
           'Principal': {'AWS': '*'},
           'Resource': bucket_resource,
           'Sid': ''}]

    if policy == Policy.READ_ONLY or policy == Policy.READ_WRITE:
        s = {'Action': list(_READ_ONLY_BUCKET_ACTIONS),
             'Effect': 'Allow',
             'Principal': {'AWS': '*'},
             'Resource': bucket_resource,
             'Sid': ''}
        if prefix:
            s.update({'Condition': {'StringEquals': {
                's3:prefix': prefix}}})
        rv.append(s)

    if policy == Policy.WRITE_ONLY or policy == Policy.READ_WRITE:
        rv.append({'Action': list(_WRITE_ONLY_BUCKET_ACTIONS),
                   'Effect': 'Allow',
                   'Principal': {'AWS': '*'},
                   'Resource': bucket_resource,
                   'Sid': ''})

    return rv


# Returns new statements contains object actions.
def _new_object_statement(policy, bucket_name, prefix=''):
    rv = [{'Action': [],
           'Effect': 'Allow',
           'Principal': {'AWS': '*'},
           'Resource': _get_object_resource(bucket_name, prefix),
           'Sid': ''}]
    if policy == Policy.READ_ONLY:
        rv[0]['Action'] = list(_READ_ONLY_OBJECT_ACTIONS)
    elif policy == Policy.WRITE_ONLY:
        rv[0]['Action'] = list(_WRITE_ONLY_OBJECT_ACTIONS)
    elif policy == Policy.READ_WRITE:
        rv[0]['Action'] = list(_READ_WRITE_OBJECT_ACTIONS)
    else:
        rv = []
    return rv


# Returns new statements for given bucket and prefix.
def _new_statements(policy, bucket_name, prefix=''):
    return (_new_bucket_statement(policy, bucket_name, prefix) +
            _new_object_statement(policy, bucket_name, prefix))


def _filter_resources(prefix, resources):
    """
    Returns a list of resources, which starts with given prefix
    """
    return list(filter(lambda r: r.startswith(prefix), resources))


# Returns whether given bucket statements are used by other than given
# prefix statements.
def _get_in_use_policy(statements, bucket_name, prefix=''):
    resource_prefix = _get_resource_prefix(bucket_name)
    object_resource = _get_object_resource(bucket_name, prefix)

    in_use = {Policy.READ_ONLY: False,
              Policy.WRITE_ONLY: False}

    for s in statements:
        resource = _get_resource(s)
        actions = set(_get_action(s))
        if (resource != object_resource and
                _filter_resources(resource_prefix, resource)):
            if (not in_use[Policy.READ_ONLY] and
                    actions & _READ_ONLY_OBJECT_ACTIONS):
                in_use[Policy.READ_ONLY] = True

            if (not in_use[Policy.WRITE_ONLY] and
                    actions & _WRITE_ONLY_OBJECT_ACTIONS):
                in_use[Policy.WRITE_ONLY] = True

    return in_use


# Removes bucket actions for given policy in given statement.
def _remove_bucket_actions(statement, policy, prefix=''):
    def _remove_read_only(s, actions, cond_value):
        if not (set(actions) & _READ_ONLY_BUCKET_ACTIONS ==
                _READ_ONLY_BUCKET_ACTIONS):
            return s, actions, cond_value

        if not prefix and not cond_value:
            actions = list(set(actions) - _READ_ONLY_BUCKET_ACTIONS)
            return s, actions, cond_value

        if prefix and cond_value:
            string_equals_value = cond_value.get('StringEquals', {})
            values = string_equals_value.get('s3:prefix', [])
            if isinstance(values, basestring):
                values = [values]

            if prefix in values:
                values.remove(prefix)

            if not values and 's3:prefix' in string_equals_value:
                del string_equals_value['s3:prefix']

            if (not string_equals_value and
                    'StringEquals' in cond_value):
                del cond_value['StringEquals']

            if not cond_value and 'Condition' in s:
                del s['Condition']

            if not cond_value:
                actions = list(set(actions) - _READ_ONLY_BUCKET_ACTIONS)
        return s, actions, cond_value

    def _remove_write_only(actions, cond_value):
        if not cond_value:
            actions = list(set(actions) - _WRITE_ONLY_BUCKET_ACTIONS)
        return actions

    actions = _get_action(statement)
    cond_value = statement.get('Condition', {})
    if (statement.get('Effect') == 'Allow' and
            statement.get('Principal') == {'AWS': '*'}):
        if policy == Policy.READ_ONLY:
            statement, actions, cond_value = _remove_read_only(statement,
                                                               actions,
                                                               cond_value)
        elif policy == Policy.WRITE_ONLY:
            actions = _remove_write_only(actions, cond_value)
        elif policy == Policy.READ_WRITE or policy == Policy.NONE:
            statement, actions, cond_value = _remove_read_only(statement,
                                                               actions,
                                                               cond_value)
            actions = _remove_write_only(actions, cond_value)

    if actions:
        statement['Action'] = actions
    else:
        statement = {}

    return statement


# Removes object actions in given statement for given policy.
def _remove_object_actions(statement, policy):
    actions = _get_action(statement)
    if (statement.get('Effect') == 'Allow' and
            statement.get('Principal') == {'AWS': '*'} and
            not statement.get('Condition')):
        if policy == Policy.READ_ONLY:
            actions = list(set(actions) - _READ_ONLY_OBJECT_ACTIONS)
        elif policy == Policy.WRITE_ONLY:
            actions = list(set(actions) - _WRITE_ONLY_OBJECT_ACTIONS)
        elif policy == Policy.READ_WRITE or policy == Policy.NONE:
            actions = list(set(actions) - _READ_WRITE_OBJECT_ACTIONS)

    if actions:
        statement['Action'] = actions
    else:
        statement = {}

    return statement


# Returns statements containing removed actions/statements for given
# policy, bucket name and prefix.
def _remove_statements(statements, policy, bucket_name, prefix=''):
    bucket_resource = _get_bucket_resource(bucket_name)
    object_resource = _get_object_resource(bucket_name, prefix)
    in_use = _get_in_use_policy(statements, bucket_name, prefix)
    out = []
    read_only_bucket_statements = []
    s3_prefix_value = []
    for s in statements:
        resource = _get_resource(s)
        if resource == bucket_resource:
            if s.get('Condition'):
                s = _remove_bucket_actions(s, policy, prefix)
            else:
                if (policy == Policy.READ_ONLY or
                        policy == Policy.READ_WRITE and
                        not in_use[Policy.READ_ONLY]):
                    s = _remove_bucket_actions(s, Policy.READ_ONLY, prefix)

                if (policy == Policy.WRITE_ONLY or
                        policy == Policy.READ_WRITE and
                        not in_use[Policy.WRITE_ONLY]):
                    s = _remove_bucket_actions(s, Policy.WRITE_ONLY, prefix)
        elif resource == object_resource:
            s = _remove_object_actions(s, policy)

        if s:
            if (resource == bucket_resource and
                    set(_get_action(s)) & _READ_ONLY_BUCKET_ACTIONS and
                    s.get('Effect') == 'Allow' and
                    s.get('Principal') == {'AWS': '*'}):
                cond_value = s.get('Condition', {})
                string_equals_value = cond_value.get('StringEquals', {})
                values = string_equals_value.get('s3:prefix', [])
                if isinstance(values, basestring):
                    values = [values]

                s3_prefix_value += [(bucket_resource[0] + '/' + v + '*')
                                    for v in values]

                if s3_prefix_value or not cond_value:
                    read_only_bucket_statements.append(s)
                    continue

            out.append(s)

    skip_bucket_statement = True
    resource_prefix = _get_resource_prefix(bucket_name)
    for s in out:
        resource = _get_resource(s)
        if (_filter_resources(resource_prefix, resource) and
                resource not in s3_prefix_value):
            skip_bucket_statement = False
            break

    for s in read_only_bucket_statements:
        if (skip_bucket_statement and
                s['Resource'] == bucket_resource and
                s.get('Effect') == 'Allow' and
                s.get('Principal') == {'AWS': '*'} and
                not s.get('Condition')):
            continue

        _append_statement(out, s)

    if len(out) == 1:
        s = out[0]
        if (s['Resource'] == bucket_resource and
                s.get('Action') == list(_COMMON_BUCKET_ACTIONS) and
                s.get('Effect') == 'Allow' and
                s.get('Principal') == {'AWS': '*'} and
                not s.get('Condition')):
            out = []

    return out


# Returns a dictionary containing values as list for same keys in
# given two dictionaries.
#
# Example:
# d1 = {'s3:prefix': 'hello',
#       's3:x-amz-acl': 'public-read',
#       'aws:SourceIp': '192.168.143.0/24'}
#
# d2 = {'s3:prefix': 'world',
#       's3:x-amz-acl': 'public-read'}
#
# _merge_dict(d1, d2) == {'s3:prefix': ['hello', 'world'],
#                         's3:x-amz-acl': ['public-read'],
#                         'aws:SourceIp': ['192.168.143.0/24']}
#
def _merge_dict(d1, d2):
    out = collections.defaultdict(list)
    for k, v in itertools.chain(d1.items(), d2.items()):
        out[k] = list(set(out[k] + [v] if isinstance(v, basestring) else v))
    return dict(out)


# Returns a dictionary containing merged conditions for given two dictionaries.
#
# Example:
# c1 = {'Condition': {'StringEquals': {'s3:prefix': 'hello'}}}
#
# c2 = {'Condition': {'StringEquals': {'s3:prefix': 'world'},
#                     'StringNotEquals': {'s3:prefix': 'foo'}}}
#
# _merge_condition(c1, c2) == {'Condition': {
#                                 'StringEquals': {
#                                     's3:prefix': ['hello', 'world']},
#                                   {'StringNotEquals': {
#                                      's3:prefix': ['foo']}}}
#
def _merge_condition(c1, c2):
    out = collections.defaultdict(dict)
    for k, v in itertools.chain(c1.items(), c2.items()):
        out.update({k:  _merge_dict(out[k], v)})
    return dict(out)


# Appends given statement into statement list to have unique statements.
# - If statement already exists in statement list, it ignores.
# - If statement exists with different conditions, they are merged.
# - Else the statement is appended to statement list.
def _append_statement(statements, statement):
    for s in statements:
        if (s.get('Resource') == statement.get('Resource') and
                _get_action(s) == _get_action(statement) and
                s.get('Effect') == statement.get('Effect') and
                s.get('Principal') == statement.get('Principal')):
            if s.get('Condition') == statement.get('Condition'):
                return
            if s.get('Condition') and statement.get('Condition'):
                s['Condition'] = _merge_condition(s['Condition'],
                                                  statement['Condition'])
                return

    statements.append(statement)


# Appends two statement lists.
def _append_statements(statements, append_statements):
    for s in append_statements:
        _append_statement(statements, s)


# Checks whether given prefix starts with given s3 prefix list.
def _match_prefix(prefix, s3_prefix):
    if isinstance(s3_prefix, basestring):
        return prefix.startswith(s3_prefix)

    matched = False
    for p in s3_prefix:
        matched = prefix.startswith(p)
        if matched:
            break

    return matched


# Returns policy of given bucket statement.
def _get_bucket_policy(statement, prefix=''):
    common_found = False
    read_only = False
    write_only = False

    actions = set(_get_action(statement))
    cond_value = statement.get('Condition', {})
    if (statement.get('Effect') == 'Allow' and
            statement.get('Principal') == {'AWS': '*'}):
        if actions & _COMMON_BUCKET_ACTIONS and not cond_value:
            common_found = True

        if actions & _WRITE_ONLY_BUCKET_ACTIONS and not cond_value:
            write_only = True

        if actions & _READ_ONLY_BUCKET_ACTIONS:
            if prefix and cond_value:
                string_equals_value = cond_value.get('StringEquals', {})
                if string_equals_value:
                    s3_prefix_value = string_equals_value.get('s3:prefix', [])
                    if isinstance(s3_prefix_value, basestring):
                        s3_prefix_value = [s3_prefix_value]
                    if _match_prefix(prefix, s3_prefix_value):
                        read_only = True
                else:
                    string_not_equals_value = cond_value.get('StringNotEquals',
                                                             {})
                    if string_not_equals_value:
                        s3_prefix_value = string_not_equals_value.get(
                            's3:prefix', [])
                        if not _match_prefix(prefix, s3_prefix_value):
                            read_only = True
            elif not prefix and not cond_value:
                read_only = True
            elif prefix and not cond_value:
                read_only = True

    return common_found, read_only, write_only


# Returns policy of given object statement.
def _get_object_policy(statement):
    read_only = False
    write_only = False

    actions = set(_get_action(statement))
    if (statement.get('Effect') == 'Allow' and
            statement.get('Principal') == {'AWS': '*'} and
            not statement.get('Condition')):
        if actions & _READ_ONLY_OBJECT_ACTIONS:
            read_only = True
        if actions & _WRITE_ONLY_OBJECT_ACTIONS:
            write_only = True

    return read_only, write_only


def _get_permissions(s, resource, object_resource, matched_resource,
                     bucket_resource, prefix, bucket_common_found,
                     bucket_read_only, bucket_write_only):

    obj_read_only, obj_write_only = False, False
    if (resource == object_resource or
            fnmatch.fnmatch(object_resource, resource)):
        read_only, write_only = _get_object_policy(s)
        if len(matched_resource) < len(resource):
            obj_read_only = read_only
            obj_write_only = write_only
            matched_resource = resource
        elif len(matched_resource) == len(resource):
            obj_read_only = obj_read_only or read_only
            obj_write_only = obj_write_only or write_only
            matched_resource = resource
    elif resource == bucket_resource:
        common_found, read_only, write_only = _get_bucket_policy(s, prefix)
        bucket_common_found = bucket_common_found or common_found
        bucket_read_only = bucket_read_only or read_only
        bucket_write_only = bucket_write_only or write_only
    return bucket_common_found, bucket_read_only, bucket_write_only, \
        obj_read_only, obj_write_only


# Returns policy of given bucket name, prefix in given statements.
def get_policy(statements, bucket_name, prefix=''):
    bucket_resource = _get_bucket_resource(bucket_name)[0]
    object_resource = _get_object_resource(bucket_name, prefix)[0]

    bucket_common_found = False
    bucket_read_only = False
    bucket_write_only = False
    matched_resource = ''
    obj_read_only = False
    obj_write_only = False
    for s in statements:
        resources = _get_resource(s)
        for resource in resources:
            bucket_common_found, bucket_read_only, bucket_write_only, \
                obj_read_only, obj_write_only = _get_permissions(
                    s, resource, object_resource, matched_resource,
                    bucket_resource, prefix, bucket_common_found,
                    bucket_read_only, bucket_write_only)

    policy = Policy.NONE
    if bucket_common_found:
        if (bucket_read_only and bucket_write_only and
                obj_read_only and obj_write_only):
            policy = Policy.READ_WRITE
        elif bucket_read_only and obj_read_only:
            policy = Policy.READ_ONLY
        elif bucket_write_only and obj_write_only:
            policy = Policy.WRITE_ONLY

    return policy


# Returns new statements containing policy of given bucket name and
# prefix are appended.
def set_policy(statements, policy, bucket_name, prefix=''):
    out = _remove_statements(statements, Policy.READ_WRITE,
                             bucket_name, prefix)
    ns = _new_statements(policy, bucket_name, prefix)
    _append_statements(out, ns)

    return out
